"""Class definition for entities that create GMHS geometry."""
from __future__ import annotations

import gmsh

from meshwell.cad import CAD


class GeometryEntity:
    """Base class for geometry entities that create GMSH geometry directly.

    Provides shared functionality for point deduplication and coordinate parsing
    to ensure consistent geometry creation across PolyLine, PolySurface, and PolyPrism.
    """

    def __init__(self, point_tolerance: float = 1e-3):
        """Initialize geometry entity with point tracking."""
        self.point_tolerance = point_tolerance
        # Track created points to avoid duplicates (can be shared across entities)
        self._points: dict[tuple[float, float, float], int] | None = None
        # Track created lines to avoid duplicates and ensure proper edge sharing
        self._lines: dict[tuple[int, int], int] | None = None

    def _parse_coords(self, coords: tuple[float, float]) -> tuple[float, float, float]:
        """Convert 2D coordinates to 3D, choosing z=0 if not provided."""
        return (coords[0], coords[1], 0) if len(coords) == 2 else coords

    def _set_point_cache(self, point_cache: dict[tuple[float, float, float], int]):
        """Set the shared point cache for this entity."""
        self._points = point_cache

    def _set_line_cache(self, line_cache: dict[tuple[int, int], int]):
        """Set the shared line cache for this entity."""
        self._lines = line_cache

    def _add_point_with_tolerance(self, x: float, y: float, z: float) -> int:
        """Add a point to the model, or reuse a previously-defined point within tolerance.

        Args:
            x: x-coordinate
            y: y-coordinate
            z: z-coordinate

        Returns:
            GMSH point ID

        """
        # Initialize local cache if no shared cache is set
        if self._points is None:
            self._points = {}

        # Snap coordinates to tolerance grid to enable point reuse
        snapped_x = round(x / self.point_tolerance) * self.point_tolerance
        snapped_y = round(y / self.point_tolerance) * self.point_tolerance
        snapped_z = round(z / self.point_tolerance) * self.point_tolerance

        key = (snapped_x, snapped_y, snapped_z)

        if key not in self._points:
            # Create new point with snapped coordinates
            self._points[key] = gmsh.model.occ.addPoint(snapped_x, snapped_y, snapped_z)

        return self._points[key]

    def _add_line_with_cache(self, point1_id: int, point2_id: int) -> int:
        """Add a line to the model, or reuse a previously-defined line between the same points.

        Args:
            point1_id: GMSH point ID #1
            point2_id: GMSH point ID #2

        Returns:
            GMSH line ID

        """
        # Initialize local cache if no shared cache is set
        if self._lines is None:
            self._lines = {}

        # Create ordered key (smaller point ID first for consistency)
        key = tuple(sorted([point1_id, point2_id]))

        if key not in self._lines:
            # Create new line
            self._lines[key] = gmsh.model.occ.addLine(point1_id, point2_id)

        return self._lines[key]

    def _create_points_from_vertices(
        self, vertices: list[tuple[float, float, float]]
    ) -> list[int]:
        """Create GMSH points from vertex coordinates, reusing existing points within tolerance.

        Args:
            vertices: List of (x, y, z) coordinates

        Returns:
            List of GMSH point IDs

        """
        points = []
        for x, y, z in vertices:
            point_id = self._add_point_with_tolerance(x, y, z)
            points.append(point_id)
        return points

    def _create_surface_from_vertices(
        self, vertices: list[tuple[float, float, float]]
    ) -> int:
        """Create a GMSH surface from vertex coordinates with point and line reuse."""
        points = self._create_points_from_vertices(vertices)

        # Create lines between consecutive points (closed loop) with caching
        lines = []
        for i in range(len(points) - 1):  # Skip last point as it should equal first
            line_id = self._add_line_with_cache(points[i], points[i + 1])
            lines.append(line_id)

        # Create closed loop and surface
        loop_id = gmsh.model.occ.addCurveLoop(lines)
        return gmsh.model.occ.addPlaneSurface([loop_id])

    def _clear_caches(self):
        """Clear the point and line caches - useful after boolean operations that may invalidate geometry."""
        if self._points is not None:
            self._points.clear()
        if self._lines is not None:
            self._lines.clear()

    def instanciate(self, cad_model: CAD | None = None) -> list[tuple[int, int]]:
        """Create GMSH geometry. To be implemented by subclasses.

        Args:
            cad_model: CAD model (kept for interface compatibility)

        Returns:
            List of (dimension, tag) tuples representing created entities

        """
        raise NotImplementedError("Subclasses must implement instanciate method")



"""Gmsh polyprism definitions."""
import gmsh
from shapely.geometry import MultiPolygon, Polygon

from meshwell.cad import CAD
from meshwell.geometry_entity import GeometryEntity
from meshwell.validation import format_physical_name


class PolyPrism(GeometryEntity):
    """Creates a bottom-up GMSH "prism" formed by a polygon associated with (optional) z-dependent grow/shrink operations.

    Attributes:
        polygons: list of shapely (Multi)Polygon
        buffers: dict of {z: buffer} used to shrink/grow base polygons at specified z-values
        physical_name: name of the physical this entity will belong to
        mesh_order: priority of the entity if it overlaps with others (lower numbers override higher numbers)
        mesh_bool: if True, entity will be meshed; if not, will not be meshed (useful to tag boundaries)

    """

    def __init__(
        self,
        polygons: Polygon | list[Polygon] | MultiPolygon | list[MultiPolygon],
        buffers: dict[float, float],
        physical_name: str | tuple[str, ...] | None = None,
        mesh_order: float | None = None,
        mesh_bool: bool = True,
        additive: bool = False,
        subdivision: tuple[int, int, int] | None = None,
        point_tolerance: float = 1e-3,
    ):
        # Initialize parent class with point tracking
        super().__init__(point_tolerance=point_tolerance)

        # Parse buffers or prepare extrusion
        if all(buffer == 0 for buffer in buffers.values()):
            self.extrude = True
            self.polygons = polygons
            self.zmin, self.zmax = min(buffers.keys()), max(buffers.keys())
        else:
            self.extrude = False
            self.buffered_polygons: list[
                tuple[float, Polygon]
            ] = self._get_buffered_polygons(polygons, buffers)

        # Store other attributes
        self.mesh_order = mesh_order
        self.additive = additive
        self.dimension = 3
        self.subdivision = subdivision

        # Format physical name
        self.physical_name = format_physical_name(physical_name)
        self.mesh_bool = mesh_bool
        self.additive = additive

    def _create_volumes_directly(self) -> list[int]:
        """Create GMSH volumes directly without using CAD class methods."""
        if self.extrude:
            surfaces = self._create_surfaces_with_holes_at_z(self.polygons, self.zmin)
            surface_dimtags = [(2, surface) for surface in surfaces]
            entities = gmsh.model.occ.extrude(
                surface_dimtags, 0, 0, self.zmax - self.zmin
            )
            volumes = [tag for dim, tag in entities if dim == 3]
        else:
            volumes = [
                self._create_volume_with_holes_directly(entry)
                for entry in self.buffered_polygons
            ]

        # Fuse multiple volumes if needed
        if len(volumes) <= 1:
            return volumes

        fused_dimtags = gmsh.model.occ.fuse(
            [(3, volumes[0])],
            [(3, tag) for tag in volumes[1:]],
            removeObject=True,
            removeTool=True,
        )[0]
        # Clear caches after boolean operations that may invalidate geometry IDs
        self._clear_caches()
        return [tag for dim, tag in fused_dimtags]

    def _get_buffered_polygons(
        self, polygons: list[Polygon], buffers: dict[float, float]
    ) -> list[tuple[float, Polygon]]:
        """Break up polygons on each layer into lists of (z,polygon) tuples according to buffer entries.

        Arguments (implicit):
            polygons: list of (Multi)Polygons to bufferize
            buffers: {z: buffer} values to apply to the polygons

        Returns:
            buffered_polygons: list of (z, buffered_polygons)

        """
        all_polygons_list = []
        for polygon in polygons.geoms if hasattr(polygons, "geoms") else [polygons]:
            current_buffers = []
            for z, width_buffer in buffers.items():
                current_buffers.append((z, polygon.buffer(width_buffer, join_style=2)))
            all_polygons_list.append(current_buffers)

        return all_polygons_list

    def _create_volume_directly(
        self,
        entry: list[tuple[float, Polygon]],
        exterior: bool = True,
        interior_index: int = 0,
    ) -> int:
        """Create volume directly using GMSH calls without CAD class methods."""
        # Draw bottom surface
        bottom_polygon = entry[0][1]
        bottom_z = entry[0][0]
        bottom_polygon_vertices = self.xy_surface_vertices(
            polygon=bottom_polygon,
            polygon_z=bottom_z,
            exterior=exterior,
            interior_index=interior_index,
        )
        gmsh_surfaces = [self._create_surface_from_vertices(bottom_polygon_vertices)]

        # Draw top surface
        top_polygon = entry[-1][1]
        top_z = entry[-1][0]
        top_polygon_vertices = self.xy_surface_vertices(
            polygon=top_polygon,
            polygon_z=top_z,
            exterior=exterior,
            interior_index=interior_index,
        )
        gmsh_surfaces.append(self._create_surface_from_vertices(top_polygon_vertices))

        # Draw vertical surfaces
        for pair_index in range(len(entry) - 1):
            if exterior:
                bottom_polygon = entry[pair_index][1].exterior.coords
                top_polygon = entry[pair_index + 1][1].exterior.coords
            else:
                bottom_polygon = entry[pair_index][1].interiors[interior_index].coords
                top_polygon = entry[pair_index + 1][1].interiors[interior_index].coords
            bottom_z = entry[pair_index][0]
            top_z = entry[pair_index + 1][0]
            for facet_pt_ind in range(len(bottom_polygon) - 1):
                facet_pt1 = (
                    bottom_polygon[facet_pt_ind][0],
                    bottom_polygon[facet_pt_ind][1],
                    bottom_z,
                )
                facet_pt2 = (
                    bottom_polygon[facet_pt_ind + 1][0],
                    bottom_polygon[facet_pt_ind + 1][1],
                    bottom_z,
                )
                facet_pt3 = (
                    top_polygon[facet_pt_ind + 1][0],
                    top_polygon[facet_pt_ind + 1][1],
                    top_z,
                )
                facet_pt4 = (
                    top_polygon[facet_pt_ind][0],
                    top_polygon[facet_pt_ind][1],
                    top_z,
                )
                facet_vertices = [facet_pt1, facet_pt2, facet_pt3, facet_pt4, facet_pt1]
                gmsh_surfaces.append(self._create_surface_from_vertices(facet_vertices))

        # Return volume from closed shell
        surface_loop = gmsh.model.occ.addSurfaceLoop(gmsh_surfaces)
        volume = gmsh.model.occ.addVolume([surface_loop])
        gmsh.model.occ.synchronize()
        return volume

    def xy_surface_vertices(
        self,
        polygon: Polygon,
        polygon_z: float,
        exterior: bool,
        interior_index: int,
    ) -> list[tuple[float, float, float]]:
        """Draw xy surface."""
        return (
            [(x, y, polygon_z) for x, y in polygon.exterior.coords]
            if exterior
            else [
                (x, y, polygon_z) for x, y in polygon.interiors[interior_index].coords
            ]
        )

    def _create_volume_with_holes_directly(
        self, entry: list[tuple[float, Polygon]]
    ) -> int:
        """Create volume with holes directly using GMSH calls."""
        exterior = self._create_volume_directly(entry, exterior=True)
        interiors = [
            self._create_volume_directly(
                entry,
                exterior=False,
                interior_index=interior_index,
            )
            for interior_index in range(len(entry[0][1].interiors))
        ]
        if interiors:
            cut_result = gmsh.model.occ.cut(
                [(3, exterior)],
                [(3, interior) for interior in interiors],
                removeObject=True,
                removeTool=True,
            )
            gmsh.model.occ.synchronize()
            # Handle cut result - if cut succeeded, use the result, otherwise keep original
            if cut_result and cut_result[0]:
                exterior = cut_result[0][0][1]  # Parse `outDimTags', `outDimTagsMap'
            # Clear caches after boolean operations that may invalidate geometry IDs
            self._clear_caches()
        return exterior

    def _create_surfaces_with_holes_at_z(self, polygons, z) -> list[int]:
        """Create surfaces with holes at given z level directly using GMSH calls."""
        surfaces = []
        for polygon in polygons.geoms if hasattr(polygons, "geoms") else [polygons]:
            # Create outer surface
            exterior_vertices = [(x, y, z) for x, y in polygon.exterior.coords]
            exterior = self._create_surface_from_vertices(exterior_vertices)

            # Create interior surfaces (holes)
            interior_surfaces = []
            for interior in polygon.interiors:
                interior_vertices = [(x, y, z) for x, y in interior.coords]
                interior_surface = self._create_surface_from_vertices(interior_vertices)
                interior_surfaces.append(interior_surface)

            # Cut holes from exterior surface
            for interior_surface in interior_surfaces:
                cut_result = gmsh.model.occ.cut(
                    [(2, exterior)],
                    [(2, interior_surface)],
                    removeObject=True,
                    removeTool=True,
                )
                gmsh.model.occ.synchronize()
                # Handle cut result - if cut succeeded, use the result, otherwise keep original
                if cut_result and cut_result[0]:
                    exterior = cut_result[0][0][
                        1
                    ]  # Parse `outDimTags', `outDimTagsMap'
                # Clear caches after boolean operations that may invalidate geometry IDs
                self._clear_caches()

            surfaces.append(exterior)
        return surfaces

    def subdivide(self, model, prisms, subdivision):
        """Split the prisms into subprisms according to subdivision."""
        subdivided_prisms = []
        import numpy as np

        global_xmin = np.inf
        global_ymin = np.inf
        global_zmin = np.inf
        global_xmax = -np.inf
        global_ymax = -np.inf
        global_zmax = -np.inf
        for prism in prisms:
            xmin, ymin, zmin, xmax, ymax, zmax = model.occ.getBoundingBox(3, prism)
            if xmin < global_xmin:
                global_xmin = xmin
            if ymin < global_ymin:
                global_ymin = ymin
            if zmin < global_zmin:
                global_zmin = zmin
            if xmax > global_xmax:
                global_xmax = xmax
            if ymax > global_ymax:
                global_ymax = ymax
            if zmax > global_zmax:
                global_zmax = zmax
        dx = (global_xmax - global_xmin) / subdivision[0]
        dy = (global_ymax - global_ymin) / subdivision[1]
        dz = (global_zmax - global_zmin) / subdivision[2]
        prisms_dimtags = {(3, prism) for prism in prisms}
        for x_index in range(subdivision[0]):
            for y_index in range(subdivision[1]):
                for z_index in range(subdivision[2]):
                    tool = model.occ.add_box(
                        global_xmin + x_index * dx,
                        global_ymin + y_index * dy,
                        global_zmin + z_index * dz,
                        dx,
                        dy,
                        dz,
                    )
                    intersection, intersection_map = model.occ.intersect(
                        list(prisms_dimtags),
                        [(3, tool)],
                        removeObject=False,
                        removeTool=True,
                    )
                    prisms_dimtags -= set(intersection_map[0])
                    subdivided_prisms.extend(intersection_map[0] + intersection)
        model.occ.remove(list(prisms_dimtags))
        return subdivided_prisms

    def instanciate(self, cad_model: CAD) -> list[tuple[int, int]]:
        """Create GMSH volumes directly without using CAD class methods."""
        prisms = self._create_volumes_directly()
        gmsh.model.occ.synchronize()
        if self.subdivision is not None:
            return self.subdivide(
                cad_model.model_manager.model, prisms, self.subdivision
            )
        # Return format consistent with original - individual prism dimtags
        return [(3, prism) for prism in prisms]

    def _validate_polygon_buffers(self) -> bool:
        """Check if any buffering operation changes the topology of the polygon."""
        # Get first polygon or multipolygon
        first_geom = self.buffered_polygons[0][0][1]

        # Handle both single polygons and multipolygons
        first_polygons = (
            first_geom.geoms if hasattr(first_geom, "geoms") else [first_geom]
        )

        # Get reference counts from first polygon(s)
        reference_counts = []
        for polygon in first_polygons:
            # Store exterior vertex count and interior vertex counts for this polygon
            polygon_counts = {
                "exterior": len(polygon.exterior.coords),
                "interiors": [len(interior.coords) for interior in polygon.interiors],
            }
            reference_counts.append(polygon_counts)

        # Check each buffered polygon matches reference counts
        for buffered_polygon in self.buffered_polygons[0][1:]:
            geom = buffered_polygon[1]
            polygons = geom.geoms if hasattr(geom, "geoms") else [geom]

            if len(polygons) != len(reference_counts):
                return False

            for polygon, ref_counts in zip(polygons, reference_counts):
                # Check exterior vertices match
                if len(polygon.exterior.coords) != ref_counts["exterior"]:
                    return False

                # Check interior vertices match
                polygon_interior_counts = [
                    len(interior.coords) for interior in polygon.interiors
                ]
                if len(polygon_interior_counts) != len(ref_counts["interiors"]):
                    return False

                for count, ref_count in zip(
                    polygon_interior_counts, ref_counts["interiors"]
                ):
                    if count != ref_count:
                        return False

        return True



"""Mesh comparison utilities."""
from difflib import unified_diff
from pathlib import Path

from meshwell.config import PATH


def compare_gmsh_files(meshfile: Path, other_meshfile: Path | None = None) -> None:
    """Compare two GMSH mesh files and assert they are identical.

    Args:
        meshfile: Path to the mesh file to compare
        other_meshfile: Optional path to compare against. If None, uses reference file
                       from PATH.references directory

    Raises:
        AssertionError: If the files differ
    """
    meshfile2 = meshfile
    if other_meshfile is None:
        meshfile1 = PATH.references / str(meshfile)
    else:
        meshfile1 = other_meshfile

    with meshfile1.open() as f:
        expected_lines = f.readlines()
    with meshfile2.open() as f:
        actual_lines = f.readlines()

    diff = list(unified_diff(expected_lines, actual_lines))
    if diff != []:
        raise ValueError(f"Mesh {meshfile!s} differs from its reference!")


def compare_mesh_headers(meshfile: Path, other_meshfile: Path | None = None) -> None:
    """Compare mesh file headers.

    Args:
        meshfile: Path to the mesh file whose header to compare
        other_meshfile: Optional path to compare against. If None, uses reference file
                       with .reference.msh extension from PATH.references directory
    """
    meshfile2 = meshfile
    if other_meshfile is None:
        meshfile1 = PATH.references / (str(meshfile.with_suffix("")) + ".reference.msh")
    else:
        meshfile1 = other_meshfile

    with meshfile1.open() as f:
        expected_lines = []
        for line in f:
            expected_lines.append(line)
            if line.startswith("$Entities"):
                # Get one more line after $Entities
                expected_lines.append(next(f))
                break

    with meshfile2.open() as f:
        actual_lines = []
        for line in f:
            actual_lines.append(line)
            if line.startswith("$Entities"):
                # Get one more line after $Entities
                actual_lines.append(next(f))
                break

    diff = list(unified_diff(expected_lines, actual_lines))
    if diff != []:
        raise ValueError(f"Mesh headers in {meshfile!s} differ from reference!")



"""PolySurface definitions."""
import gmsh
from shapely.geometry import MultiPolygon, Polygon

from meshwell.cad import CAD
from meshwell.geometry_entity import GeometryEntity


class PolySurface(GeometryEntity):
    """Creates bottom-up GMSH polygonal surfaces formed by list of shapely (multi)polygon.

    Attributes:
        polygons: list of shapely (Multi)Polygon
        model: GMSH model to synchronize
        physical_name: name of the physical this entity will belong to
        mesh_order: priority of the entity if it overlaps with others (lower numbers override higher numbers)

    """

    def __init__(
        self,
        polygons: Polygon | list[Polygon] | MultiPolygon | list[MultiPolygon],
        physical_name: str | tuple[str, ...] | None = None,
        mesh_order: float | None = None,
        mesh_bool: bool = True,
        additive: bool = False,
        point_tolerance: float = 1e-3,
    ):
        # Initialize parent class with point tracking
        super().__init__(point_tolerance=point_tolerance)

        # Parse (multi)polygons
        if isinstance(polygons, (Polygon, MultiPolygon)):
            self.polygons = list(
                polygons.geoms if hasattr(polygons, "geoms") else [polygons]
            )
        else:
            self.polygons = []
            for entry in polygons:
                self.polygons.extend(
                    entry.geoms if hasattr(entry, "geoms") else [entry]
                )

        self.mesh_order = mesh_order
        if isinstance(physical_name, str):
            self.physical_name = (physical_name,)
        else:
            self.physical_name = physical_name
        self.mesh_bool = mesh_bool
        self.dimension = 2
        self.additive = additive

    def _create_surface_with_holes(self, polygon: Polygon) -> int:
        """Create surface with holes directly using GMSH calls."""
        # Create exterior surface
        exterior_vertices = [
            self._parse_coords(coords) for coords in polygon.exterior.coords
        ]
        exterior = self._create_surface_from_vertices(exterior_vertices)

        # Create interior surfaces (holes)
        interior_surfaces = []
        for interior in polygon.interiors:
            interior_vertices = [
                self._parse_coords(coords) for coords in interior.coords
            ]
            interior_surface = self._create_surface_from_vertices(interior_vertices)
            interior_surfaces.append(interior_surface)

        # Cut holes from exterior surface
        for interior_surface in interior_surfaces:
            cut_result = gmsh.model.occ.cut(
                [(2, exterior)],
                [(2, interior_surface)],
                removeObject=True,
                removeTool=True,
            )
            gmsh.model.occ.synchronize()
            exterior = cut_result[0][0][1]  # Parse `outDimTags', `outDimTagsMap'
            # Clear caches after boolean operations that may invalidate geometry IDs
            self._clear_caches()

        return exterior

    def instanciate(
        self,
        cad_model: CAD | None = None,  # noqa: ARG002
    ) -> list[tuple[int, int]]:
        """Create GMSH surfaces directly without using CAD class methods."""
        surfaces = []
        for polygon in self.polygons:
            surface_id = self._create_surface_with_holes(polygon)
            surfaces.append(surface_id)

        # Fuse multiple surfaces if needed
        if len(surfaces) <= 1:
            gmsh.model.occ.synchronize()
            return [(2, surfaces[0])] if surfaces else []

        # Fuse all surfaces together
        dimtags = gmsh.model.occ.fuse(
            [(2, surfaces[0])],
            [(2, tag) for tag in surfaces[1:]],
            removeObject=True,
            removeTool=True,
        )[0]
        gmsh.model.occ.synchronize()
        # Clear caches after boolean operations that may invalidate geometry IDs
        self._clear_caches()
        return dimtags



"""Gmsh wire definitions."""
import gmsh
from shapely.geometry import LineString, MultiLineString

from meshwell.cad import CAD
from meshwell.geometry_entity import GeometryEntity


class PolyLine(GeometryEntity):
    """Creates bottom-up GMSH wires formed by list of shapely (multi)linestring.

    Attributes:
        linestrings: list of shapely (Multi)LineString
        physical_name: name of the physical this entity will belong to
        mesh_order: priority of the entity if it overlaps with others (lower numbers override higher numbers)

    """

    def __init__(
        self,
        linestrings: LineString
        | list[LineString]
        | MultiLineString
        | list[MultiLineString],
        physical_name: str | tuple[str, ...] | None = None,
        mesh_order: float | None = None,
        mesh_bool: bool = True,
        additive: bool = False,
        point_tolerance: float = 1e-3,
    ):
        # Initialize parent class with point tracking
        super().__init__(point_tolerance=point_tolerance)

        # Parse (multi)linestrings
        if isinstance(linestrings, list):
            # Handle list of LineString/MultiLineString objects
            self.linestrings = []
            for item in linestrings:
                if hasattr(item, "geoms"):  # MultiLineString
                    self.linestrings.extend(list(item.geoms))
                else:  # LineString
                    self.linestrings.append(item)
        elif hasattr(linestrings, "geoms"):  # Single MultiLineString
            self.linestrings = list(linestrings.geoms)
        else:  # Single LineString
            self.linestrings = [linestrings]

        self.mesh_order = mesh_order
        if isinstance(physical_name, str):
            self.physical_name = (physical_name,)
        else:
            self.physical_name = physical_name
        self.mesh_bool = mesh_bool
        self.dimension = 1
        self.additive = additive

    def _create_wire_from_linestring(self, linestring: LineString) -> int:
        """Create a GMSH wire directly from linestring coordinates."""
        vertices = [self._parse_coords(coords) for coords in linestring.coords]

        # Create points with deduplication
        points = self._create_points_from_vertices(vertices)

        # Create lines between consecutive points
        lines = []
        for i in range(len(points) - 1):
            line_id = gmsh.model.occ.addLine(points[i], points[i + 1])
            lines.append(line_id)

        # Create wire from lines
        if len(lines) == 1:
            # For a single line, we can return it as-is since GMSH treats it as a wire
            return lines[0]
        # For multiple lines, create a proper wire
        return gmsh.model.occ.addWire(lines)

    def instanciate(
        self,
        cad_model: CAD | None = None,  # noqa: ARG002
    ) -> list[tuple[int, int]]:
        """Create GMSH wires directly without using CAD class methods."""
        wires = []
        for linestring in self.linestrings:
            wire_id = self._create_wire_from_linestring(linestring)
            wires.append(wire_id)

        gmsh.model.occ.synchronize()
        return [(1, wire) for wire in wires]
